import datetime
import pytz

from mock import MagicMock


class IsInstanceMatcher:
    def __init__(self, expected_type):
        self.expected_type = expected_type

    def __eq__(self, other):
        return isinstance(other, self.expected_type)


class IsInstanceMockMatcher:
    def __init__(self, expected_type):
        self.expected_type = expected_type

    def __eq__(self, other):
        return isinstance(other, MagicMock) and other.type == str(self.expected_type)


class GeneratorMock:
    def __init__(self, data):
        self.data = data
        self.time_extracted = datetime.datetime.utcnow().replace(tzinfo=pytz.UTC)
        self.endpoint_bookmark_field = "last_modified_date"
        self.endpoint_sorting_criteria = {'field': 'id'}
        self.client = "mock_client"
        self.state = "mock_state"
        self.config = "mock_config"

    def __iter__(self):
        return self

    def __next__(self):
        if not self.data:
            raise StopIteration
        return self.data.pop(0)

    def write_bookmark(self):
        pass


import singer

from .tap_generators.deposit_accounts_generator import DepositAccountsGenerator
from .tap_generators.deposit_cards_generator import DepositCardsGenerator
from .tap_generators.loan_accounts_generator import LoanAccountsADGenerator, LoanAccountsLMGenerator
from .tap_generators.loan_repayments_generator import LoanRepaymentsGenerator
from .tap_processors.deposit_accounts_processor import DepositAccountsProcessor
from .tap_processors.deposit_cards_processor import DepositCardsProcessor
from .tap_processors.loan_accounts_processor import LoanAccountsProcessor
from .tap_processors.loan_repayments_processor import LoanRepaymentsProcessor

LOGGER = singer.get_logger()


def sync_endpoint_refactor(client, catalog, state,
                           stream_name, sub_type, config, parent_id=None):
    stream_generator_processor_dict = {
        "loan_accounts": ((LoanAccountsLMGenerator, LoanAccountsADGenerator), LoanAccountsProcessor),
        "loan_repayments": ((LoanRepaymentsGenerator,), LoanRepaymentsProcessor),
        "deposit_accounts": ((DepositAccountsGenerator,), DepositAccountsProcessor),
        "cards": ((DepositCardsGenerator,), DepositCardsProcessor)
    }

    generator_classes, processor_class = stream_generator_processor_dict[stream_name]
    generators = [generator_class(stream_name=stream_name,
                                  client=client,
                                  config=config,
                                  state=state,
                                  sub_type=sub_type,
                                  **{} if parent_id is None else {"parent_id": parent_id})
                  for generator_class in generator_classes]
    processor = processor_class(catalog=catalog,
                                stream_name=stream_name,
                                client=client,
                                config=config,
                                state=state,
                                sub_type=sub_type,
                                generators=generators,
                                **{} if parent_id is None else {"parent_id": parent_id})

    return processor.process_streams_from_generators()



from .child_generator import ChildGenerator


class DepositCardsGenerator(ChildGenerator):
    def _init_endpoint_config(self):
        super(DepositCardsGenerator, self)._init_endpoint_config()
        self.endpoint_api_method = "GET"
        self.endpoint_path = f"deposits/{self.endpoint_parent_id}/cards"

    def _init_params(self):
        self.time_extracted = None

    def prepare_batch(self):
        self.params = {}

    def __next__(self):
        if not self.buffer:
            raise StopIteration()
        return self.buffer.pop(0)



from mock import MagicMock

from ..constants import config_json


def test_deposit_accounts_generator():
    from tap_mambu.tap_mambu_refactor.tap_generators.deposit_accounts_generator import DepositAccountsGenerator
    client_mock = MagicMock()
    client_mock.page_size = int(config_json.get("page_size", 500))
    client_mock.request = MagicMock()
    generator = DepositAccountsGenerator(stream_name="deposit_accounts",
                                            client=client_mock,
                                            config=config_json,
                                            state={'currently_syncing': 'deposit_accounts'},
                                            sub_type="self")
    assert generator.endpoint_path == "deposits:search"
    assert generator.endpoint_bookmark_field == "lastModifiedDate"
    assert generator.endpoint_sorting_criteria == {
            "field": "lastModifiedDate",
            "order": "ASC"
        }
    assert generator.endpoint_filter_criteria == [
            {
                "field": "lastModifiedDate",
                "operator": "AFTER",
                "value": '2021-06-01'
            }
        ]



from .processor import TapProcessor, LOGGER
from ..helpers import get_selected_streams


class ParentProcessor(TapProcessor):
    def _process_child_records(self, record):
        from .. import sync_endpoint_refactor

        total_records = super()._process_child_records(record)
        for child_stream_name in self.endpoint_child_streams:
            if child_stream_name in get_selected_streams(self.catalog):
                parent_id = record[self.endpoint_id_field]
                LOGGER.info(f'(processor) Syncing: {child_stream_name}, parent_stream: {self.stream_name}, parent_id: {parent_id}')

                child_synced_records = sync_endpoint_refactor(client=self.client,
                                                              catalog=self.catalog,
                                                              state=self.state,
                                                              stream_name=child_stream_name,
                                                              sub_type='self',
                                                              config=self.config,
                                                              parent_id=parent_id)
                total_records += child_synced_records
                LOGGER.info(f'(processor) Synced: {child_stream_name}, parent_id: {parent_id}, synced records: {child_synced_records}')
        return total_records



from .parent_processor import ParentProcessor


class LoanAccountsProcessor(ParentProcessor):
    def _init_endpoint_config(self):
        super(LoanAccountsProcessor, self)._init_endpoint_config()
        self.endpoint_deduplication_key = "id"
        self.endpoint_child_streams = ["loan_repayments"]



from .parent_processor import ParentProcessor


class DepositAccountsProcessor(ParentProcessor):
    def _init_endpoint_config(self):
        super(DepositAccountsProcessor, self)._init_endpoint_config()
        self.endpoint_deduplication_key = "id"
        self.endpoint_child_streams = ["cards"]



import mock
from mock import MagicMock

from ..constants import config_json


def test_loan_accounts_lm_generator():
    from tap_mambu.tap_mambu_refactor.tap_generators.loan_accounts_generator import LoanAccountsLMGenerator
    client_mock = MagicMock()
    client_mock.page_size = int(config_json.get("page_size", 500))
    client_mock.request = MagicMock()
    lm_generator = LoanAccountsLMGenerator(stream_name="loan_accounts",
                                           client=client_mock,
                                           config=config_json,
                                           state={'currently_syncing': 'loan_accounts'},
                                           sub_type="self")
    assert lm_generator.endpoint_path == "loans:search"
    assert lm_generator.endpoint_bookmark_field == "lastModifiedDate"
    assert lm_generator.endpoint_sorting_criteria == {
            "field": "id",
            "order": "ASC"
        }
    assert lm_generator.endpoint_filter_criteria == [
            {
                "field": "lastModifiedDate",
                "operator": "AFTER",
                "value": '2021-06-01'
            }
        ]


def test_loan_accounts_ad_generator():
    client_mock = MagicMock()
    client_mock.page_size = int(config_json.get("page_size", 500))
    client_mock.request = MagicMock()
    from tap_mambu.tap_mambu_refactor.tap_generators.loan_accounts_generator import LoanAccountsADGenerator
    ad_generator = LoanAccountsADGenerator(stream_name="loan_accounts",
                                           client=client_mock,
                                           config=config_json,
                                           state={"currently_syncing": "loan_accounts"},
                                           sub_type="self")
    assert ad_generator.endpoint_path == "loans:search"
    assert ad_generator.endpoint_bookmark_field == "lastAccountAppraisalDate"
    assert ad_generator.endpoint_sorting_criteria == {
            "field": "id",
            "order": "ASC"
        }
    assert ad_generator.endpoint_filter_criteria == [
        {
            "field": "lastAccountAppraisalDate",
            "operator": "AFTER",
            "value": '2021-06-01'
        }
    ]



from unittest.mock import MagicMock

import mock
import tap_mambu.tap_mambu_refactor.tap_generators.child_generator

from ..constants import config_json


def test_repayments_generator_endpoint_config_init():
    from tap_mambu.tap_mambu_refactor.tap_generators.loan_repayments_generator import LoanRepaymentsGenerator
    client_mock = MagicMock()
    client_mock.page_size = int(config_json.get("page_size", 500))
    client_mock.request = MagicMock()
    generator = LoanRepaymentsGenerator(stream_name="loan_accounts",
                                        client=client_mock,
                                        config=config_json,
                                        state={'currently_syncing': 'loan_accounts'},
                                        sub_type="self",
                                        parent_id="ASD")

    assert generator.endpoint_api_key_type is None
    assert generator.endpoint_parent_id == "ASD"
    assert generator.endpoint_api_version == "v1"



import inspect
import os

from mock import MagicMock

from ..constants import config_json
from ..helpers import GeneratorMock

FIXTURES_PATH = f"{os.path.dirname(os.path.abspath(inspect.stack()[0][1]))}/Fixtures"


def test_deposit_cards_processor_endpoint_config_init():
    from singer.catalog import Catalog
    catalog = Catalog.load(f"{FIXTURES_PATH}/processor_catalog.json")
    client_mock = MagicMock()

    from tap_mambu.tap_mambu_refactor.tap_processors.deposit_cards_processor import DepositCardsProcessor
    processor = DepositCardsProcessor(catalog=catalog,
                                      stream_name="cards",
                                      client=client_mock,
                                      config=config_json,
                                      state={'currently_syncing': 'cards'},
                                      sub_type="self",
                                      generators=[GeneratorMock([])],
                                      parent_id='TEST')

    assert processor.endpoint_deduplication_key == 'reference_token'
    assert processor.endpoint_id_field == 'reference_token'
    assert processor.endpoint_parent == 'deposit'
    assert processor.endpoint_parent_id == 'TEST'



import inspect
import json
import os
import mock
from unittest.mock import MagicMock

from tap_mambu.tap_mambu_refactor.tap_generators.generator import TapGenerator
from ..constants import config_json

FIXTURES_PATH = f"{os.path.dirname(os.path.abspath(inspect.stack()[0][1]))}/fixtures"


# def test_count_records():
#     client_mock = MagicMock()
#     client_mock.page_size = int(config_json.get("page_size", 500))
#     client_mock.request = MagicMock()
#     with open(f"{FIXTURES_PATH}/data_LoanAccountsLMGenerator.json", "r") as fd:
#         fixture_data = json.loads(fd.read())
#     client_mock.request.side_effect = fixture_data
#     generator = TapGenerator(stream_name="loan_accounts",
#                              client=client_mock,
#                              config=config_json,
#                              state={"currently_syncing": "loan_accounts"},
#                              sub_type="self")
#
#     record_count = 0
#     for _ in generator:
#         record_count += 1
#
#     assert record_count == sum([len(record_list) for record_list in fixture_data])


def test_transform_batch_custom_field():
    from tap_mambu.tap_mambu_refactor.helpers import transform_json

    assert transform_json([
        {"encodedKey": "123", "creation_date": "asd",
         "_random_field": {"cevaId": "asd"},
         "_random_field_2": [{"a": "b", "c": "d"}],
         "_random_field_3": 0}
    ], "loan_accounts") == [
        {"encoded_key": "123", "creation_date": "asd", "custom_fields": [
            {"field_set_id": "_random_field", "id": "cevaId", "value": "asd"},
            {"field_set_id": "_random_field_2", "id": "a", "value": "b"},
            {"field_set_id": "_random_field_2", "id": "c", "value": "d"}
        ]}
    ]


def test_transform_batch_custom_field_auto_add():
    from tap_mambu.tap_mambu_refactor.helpers import transform_json

    assert transform_json([
        dict(encodedKey="123", creation_date="asd", _test=2)
    ], "loan_accounts") == [
        dict(encoded_key="123", creation_date="asd", custom_fields=list())
    ]


def test_transform_batch_larger_batch():
    from tap_mambu.tap_mambu_refactor.helpers import transform_json

    assert transform_json([
        dict(encodedKey="123", creation_date="asd", custom_fields=list()) for _ in range(10)
    ], "loan_accounts") == [
        dict(encoded_key="123", creation_date="asd", custom_fields=list()) for _ in range(10)
    ]



import mock
from mock import PropertyMock

from tap_mambu.tap_mambu_refactor import sync_endpoint_refactor
from .helpers import IsInstanceMockMatcher


@mock.patch("tap_mambu.tap_mambu_refactor.main.LoanAccountsLMGenerator")
@mock.patch("tap_mambu.tap_mambu_refactor.main.LoanAccountsADGenerator")
@mock.patch("tap_mambu.tap_mambu_refactor.main.LoanAccountsProcessor")
def test_sync_endpoint_refactor(mock_loan_accounts_processor, mock_loan_accounts_ad_generator,
                                mock_loan_accounts_lm_generator):
    type(mock_loan_accounts_lm_generator.return_value).type = PropertyMock(return_value="LoanAccountsLMGenerator")
    type(mock_loan_accounts_ad_generator.return_value).type = PropertyMock(return_value="LoanAccountsADGenerator")
    client = "client"
    catalog = "catalog"
    state = "state"
    stream_name = "loan_accounts"
    sub_type = "sub_type"
    config = "config"
    sync_endpoint_refactor(client=client, catalog=catalog, state=state, stream_name=stream_name,
                           sub_type=sub_type, config=config)
    mock_loan_accounts_lm_generator.assert_called_once_with(client=client, state=state, stream_name=stream_name,
                                                            sub_type=sub_type, config=config)
    mock_loan_accounts_ad_generator.assert_called_once_with(client=client, state=state, stream_name=stream_name,
                                                            sub_type=sub_type, config=config)
    mock_loan_accounts_processor.assert_called_once_with(client=client, state=state, stream_name=stream_name,
                                                         sub_type=sub_type, config=config, catalog=catalog,
                                                         generators=[IsInstanceMockMatcher("LoanAccountsLMGenerator"),
                                                                     IsInstanceMockMatcher("LoanAccountsADGenerator")])



import inspect
import os

from mock import MagicMock

from ..constants import config_json
from ..helpers import GeneratorMock

FIXTURES_PATH = f"{os.path.dirname(os.path.abspath(inspect.stack()[0][1]))}/Fixtures"


def test_loan_repayments_processor_endpoint_config_init():
    from singer.catalog import Catalog
    catalog = Catalog.load(f"{FIXTURES_PATH}/processor_catalog.json")
    client_mock = MagicMock()

    from tap_mambu.tap_mambu_refactor.tap_processors.loan_repayments_processor import LoanRepaymentsProcessor
    processor = LoanRepaymentsProcessor(catalog=catalog,
                                        stream_name="loan_repayments",
                                        client=client_mock,
                                        config=config_json,
                                        state={'currently_syncing': 'loan_repayments'},
                                        sub_type="self",
                                        generators=[GeneratorMock([])],
                                        parent_id='TEST')

    assert processor.endpoint_parent == 'loan_accounts'
    assert processor.endpoint_parent_id == 'TEST'



from unittest.mock import MagicMock

from ..constants import config_json


def test_deposit_cards_generator_endpoint_config_init():
    from tap_mambu.tap_mambu_refactor.tap_generators.deposit_cards_generator import DepositCardsGenerator
    client_mock = MagicMock()
    client_mock.page_size = int(config_json.get("page_size", 500))
    client_mock.request = MagicMock()
    generator = DepositCardsGenerator(stream_name="deposit_accounts",
                                      client=client_mock,
                                      config=config_json,
                                      state={'currently_syncing': 'deposit_accounts'},
                                      sub_type="self",
                                      parent_id="TEST")

    assert generator.endpoint_parent_id == "TEST"
    assert generator.endpoint_path == 'deposits/TEST/cards'
    assert generator.endpoint_api_method == "GET"



import inspect
import os

from mock import MagicMock

from ..constants import config_json
from ..helpers import GeneratorMock

FIXTURES_PATH = f"{os.path.dirname(os.path.abspath(inspect.stack()[0][1]))}/Fixtures"


def test_loan_accounts_processor():
    from singer.catalog import Catalog
    catalog = Catalog.load(f"{FIXTURES_PATH}/processor_catalog.json")
    client_mock = MagicMock()

    from tap_mambu.tap_mambu_refactor.tap_processors.loan_accounts_processor import LoanAccountsProcessor
    processor = LoanAccountsProcessor(catalog=catalog,
                                      stream_name="loan_accounts",
                                      client=client_mock,
                                      config=config_json,
                                      state={'currently_syncing': 'loan_accounts'},
                                      sub_type="self",
                                      generators=[GeneratorMock([])])

    assert processor.endpoint_deduplication_key == "id"
    assert processor.endpoint_child_streams == ["loan_repayments"]


import datetime
import json

import pytest
from mock import MagicMock, call

import mock
import os.path
import inspect

from ..constants import config_json
from ..helpers import GeneratorMock, IsInstanceMatcher

FIXTURES_PATH = f"{os.path.dirname(os.path.abspath(inspect.stack()[0][1]))}/Fixtures"


# @mock.patch("tap_mambu.tap_mambu_refactor.tap_processors.processor.TapProcessor.write_schema")
# @mock.patch("tap_mambu.tap_mambu_refactor.tap_processors.processor.TapProcessor.write_bookmark")
# def test_tap_processor(mock_write_bookmark, mock_write_schema, capsys):
#     from singer.catalog import Catalog
#     from tap_mambu.tap_mambu_refactor.tap_processors.processor import TapProcessor
#     catalog = Catalog.load(f"{FIXTURES_PATH}/catalog.json")
#
#     with open(f"{FIXTURES_PATH}/data_LoanAccountsLMGenerator.json", "r") as fd:
#         loan_accounts_lm_generator = GeneratorMock(json.loads(fd.read()))
#         loan_accounts_lm_generator.time_extracted = 0
#
#     with open(f"{FIXTURES_PATH}/data_LoanAccountsADGenerator.json", "r") as fd:
#         loan_accounts_ad_generator = GeneratorMock(json.loads(fd.read()))
#         loan_accounts_ad_generator.time_extracted = 0
#
#     client_mock = MagicMock()
#     processor = TapProcessor(catalog=catalog,
#                              stream_name="loan_accounts",
#                              client=client_mock,
#                              config=config_json,
#                              state={'currently_syncing': 'loan_accounts'},
#                              sub_type="self",
#                              generators=[loan_accounts_lm_generator, loan_accounts_ad_generator])
#     processor.endpoint_deduplication_key = "id"
#     processor.process_streams_from_generators()
#
#     captured = capsys.readouterr()
#
#     stdout_list = [json.loads(line) for line in captured.out.split("\n") if line]
#     with open(f"{FIXTURES_PATH}/expected_LoanAccountsProcessor.json", "r") as fd:
#         expected_list = [dict(type="RECORD", stream="loan_accounts", record=record) for record in json.loads(fd.read())]
#
#     for record in stdout_list:
#         assert record in expected_list
#
#     for expected_record in expected_list:
#         assert expected_record in stdout_list


@mock.patch("tap_mambu.tap_mambu_refactor.tap_processors.parent_processor.get_selected_streams")
@mock.patch("tap_mambu.tap_mambu_refactor.tap_processors.processor.TapProcessor.write_schema")
@mock.patch("tap_mambu.tap_mambu_refactor.tap_processors.processor.TapProcessor.write_bookmark")
@mock.patch("tap_mambu.tap_mambu_refactor.sync_endpoint_refactor")
def test_tap_processor_process_child_records(mock_sync_endpoint_refactor,
                                             mock_write_bookmark,
                                             mock_write_schema,  # Mock write_schema so we don't pollute the output
                                             mock_get_selected_streams,
                                             capsys):
    from singer.catalog import Catalog
    from tap_mambu.tap_mambu_refactor.tap_processors.parent_processor import ParentProcessor
    fake_children_record_count = 4
    mock_get_selected_streams.return_value = ["child_1", "child_2"]
    mock_sync_endpoint_refactor.return_value = fake_children_record_count
    catalog = Catalog.load(f"{FIXTURES_PATH}/processor_catalog.json")

    generator_data = [
        {
            "encoded_key": "12345678901234567890123456789012",
            "last_modified_date": "2022-01-01T00:00:00.000000Z",
            "id": "2"
        }, {
            "encoded_key": "12345678901234567890123456789013",
            "last_modified_date": "2022-01-01T00:00:00.000000Z",
            "id": "3"
        }, {
            "encoded_key": "12345678901234567890123456789014",
            "last_modified_date": "2022-01-01T00:00:00.000000Z",
            "id": "4"
        }, {
            "encoded_key": "12345678901234567890123456789015",
            "last_modified_date": "2022-01-01T00:00:00.000000Z",
            "id": "5"
        }
    ]
    generator = GeneratorMock(list(generator_data))
    generator.time_extracted = 0

    client_mock = MagicMock()
    processor = ParentProcessor(catalog=catalog,
                                stream_name="loan_accounts",
                                client=client_mock,
                                config=config_json,
                                state={'currently_syncing': 'loan_accounts'},
                                sub_type="self",
                                generators=[generator])
    processor.endpoint_child_streams = ["child_1", "child_2"]
    actual_records_count = processor.process_streams_from_generators()
    # sync_endpoint_refactor called for every record (len(generator_data)) for every child_stream + once for parent
    assert actual_records_count == \
           len(generator_data) * (fake_children_record_count * len(processor.endpoint_child_streams) + 1), \
        "Record count mismatch when adding child records"

    mock_sync_endpoint_refactor.assert_called_with(client=client_mock,
                                                   catalog=processor.catalog,
                                                   state={'currently_syncing': 'loan_accounts'},
                                                   stream_name=processor.endpoint_child_streams[-1],
                                                   sub_type="self",
                                                   config=config_json,
                                                   parent_id="5")

    captured = capsys.readouterr()
    stdout_list = [json.loads(line) for line in captured.out.split("\n") if line]
    # noinspection PyTypeChecker
    assert stdout_list == [
        {"type": "RECORD", "stream": "loan_accounts", "record": record} for record in generator_data
    ], "Output should contain mocked records"


@mock.patch("tap_mambu.tap_mambu_refactor.helpers.write_state")
def test_bookmarks(mock_write_state):
    from singer.catalog import Catalog
    from tap_mambu.tap_mambu_refactor.tap_processors.processor import TapProcessor

    catalog = Catalog.load(f"{FIXTURES_PATH}/processor_catalog.json")
    client_mock = MagicMock()
    processor = TapProcessor(catalog=catalog,
                             stream_name="loan_accounts",
                             client=client_mock,
                             config=config_json,
                             state={'currently_syncing': 'loan_accounts'},
                             sub_type="self",
                             generators=[GeneratorMock([])])

    processor.write_bookmark()

    expected_state = {'currently_syncing': 'loan_accounts', 'bookmarks': {
        'loan_accounts': '2021-06-01T00:00:00Z'}}
    mock_write_state.assert_called_once_with(expected_state)


@mock.patch("tap_mambu.tap_mambu_refactor.tap_processors.processor.write_schema")
def test_write_schema(mock_write_schema):
    from singer.catalog import Catalog
    from tap_mambu.tap_mambu_refactor.tap_processors.processor import TapProcessor

    catalog = Catalog.load(f"{FIXTURES_PATH}/processor_catalog.json")
    client_mock = MagicMock()
    processor = TapProcessor(catalog=catalog,
                             stream_name="loan_accounts",
                             client=client_mock,
                             config=config_json,
                             state={'currently_syncing': 'loan_accounts'},
                             sub_type="self",
                             generators=[GeneratorMock([])])
    processor.write_schema()

    schema = None
    stream_key_properties = None
    with open(f"{FIXTURES_PATH}/processor_catalog.json", "r") as fd:
        schema_json = json.loads(fd.read())
        for stream in schema_json["streams"]:
            if stream["tap_stream_id"] == "loan_accounts":
                schema = stream["schema"]
                stream_key_properties = stream["key_properties"]
                break
    assert schema is not None
    mock_write_schema.assert_called_with("loan_accounts", schema, stream_key_properties)


@mock.patch("tap_mambu.tap_mambu_refactor.tap_processors.processor.write_record")
@mock.patch("tap_mambu.tap_mambu_refactor.tap_processors.processor.write_schema")
def test_write_exceptions(mock_write_schema, mock_write_record):
    from singer.catalog import Catalog
    from tap_mambu.tap_mambu_refactor.tap_processors.processor import TapProcessor
    catalog = Catalog.load(f"{FIXTURES_PATH}/processor_catalog.json")

    mock_write_record.side_effect = [None, OSError("Mock Record Exception")]
    mock_write_schema.side_effect = [None, OSError("Mock Schema Exception")]

    client_mock = MagicMock()
    processor = TapProcessor(catalog=catalog,
                             stream_name="loan_accounts",
                             client=client_mock,
                             config=config_json,
                             state={'currently_syncing': 'loan_accounts'},
                             sub_type="self",
                             generators=[GeneratorMock([
                                 {"id": "1", "last_modified_date": "2022-01-01T00:00:00+03:00"},
                                 {"id": "2", "last_modified_date": "2022-01-01T01:00:00+03:00"},
                                 {"id": "3", "last_modified_date": "2022-01-01T02:00:00+03:00"}
                             ])])
    processor.endpoint_deduplication_key = "id"

    with pytest.raises(OSError) as err:
        processor.process_streams_from_generators()
    assert err.value.args[0] == "Mock Record Exception"

    mock_write_record.assert_has_calls([
        call("loan_accounts", IsInstanceMatcher(dict), time_extracted=IsInstanceMatcher(datetime.datetime)),
        call("loan_accounts", IsInstanceMatcher(dict), time_extracted=IsInstanceMatcher(datetime.datetime))
    ])

    with pytest.raises(OSError) as err:
        processor.process_streams_from_generators()
    assert err.value.args[0] == "Mock Schema Exception"

    mock_write_schema.assert_has_calls([
        call("loan_accounts", IsInstanceMatcher(dict), IsInstanceMatcher(list)),
        call("loan_accounts", IsInstanceMatcher(dict), IsInstanceMatcher(list))
    ])
